/*
 * (c) 2008-2009 Adam Lackorzynski <adam@os.inf.tu-dresden.de>
 *     economic rights: Technische Universität Dresden (Germany)
 * Copyright (C) 2017, 2018, 2021 Paul Boddie <paul@boddie.org.uk>
 *
 * This file is part of TUD:OS and distributed under the terms of the
 * GNU General Public License 2.
 * Please see the COPYING-GPL-2 file for details.
 */
/*
 * Access the clock and power management unit on the MIPS Creator CI20 board.
 */

#include <l4/devices/cpm-jz4780.h>
#include <l4/io/io.h>
#include <l4/re/env.h>
#include <l4/re/c/util/cap_alloc.h>
#include <l4/sys/factory.h>
#include <l4/sys/icu.h>
#include <l4/sys/ipc.h>
#include <l4/sys/irq.h>
#include <l4/vbus/vbus.h>
#include <stdio.h>
#include <unistd.h>
#include <stdint.h>



/* Device and resource discovery. */

static char const *resource_type(enum l4io_resource_types_t type)
{
  switch (type)
  {
    case L4VBUS_RESOURCE_INVALID:
    return "INVALID";

    case L4VBUS_RESOURCE_IRQ:
    return "IRQ";

    case L4VBUS_RESOURCE_MEM:
    return "MEMORY";

    default:
    return "OTHER";
  }
}

static int cpm_get_device(char const *hid, l4io_device_handle_t *dh, l4io_resource_handle_t *rh)
{
  int result = l4io_lookup_device(hid, dh, 0, rh);

  if (result < 0)
    printf("Could not access '%s': %s\n", hid, result == -L4_ENOENT ? "no such device" : "no device");

  return result;
}

static int cpm_get_resource(l4io_device_handle_t dh, l4io_resource_t *res,
                            enum l4io_resource_types_t type)
{
  int current = 0, result = 0;
  l4_cap_idx_t vbus = l4re_env_get_cap("vbus");

  do
  {
    result = l4vbus_get_resource(vbus, dh, current, res);

    if (result)
      printf("Could not access resource of type %s.\n", resource_type(type));
    else
      printf("Resource %d: type %s, start=%lx, end=%lx\n", res->id,
        resource_type(res->type), res->start, res->end);

    current++;
  }
  while ((!result) && (res->type != type));

  return result;
}

static int cpm_get_memory(char const *hid, l4_addr_t *start, l4_addr_t *end)
{
  l4io_device_handle_t dh;
  l4io_resource_handle_t rh;
  l4io_resource_t res;
  int result;

  result = cpm_get_device(hid, &dh, &rh);

  if (result < 0)
    return result;

  result = cpm_get_resource(dh, &res, L4IO_RESOURCE_MEM);

  if (result)
    return result;

  if ((result = l4io_request_iomem(res.start, res.end - res.start + 1,
                                  L4IO_MEM_NONCACHED, start)))
  {
    printf("Could not get address for '%s'.\n", hid);
    return result;
  }

  printf("Resource at 0x%lx...0x%lx.\n", res.start, res.end);

  *end = *start + (res.end - res.start + 1);

  return 0;
}

int main(void)
{
  l4_addr_t cpm_base = 0, cpm_base_end = 0;
  void *cpm;
  int result = 0;

  /* Obtain resource details describing I/O memory. */

  printf("Access CPM...\n");

  if ((result = cpm_get_memory("jz4780-cpm", &cpm_base, &cpm_base_end)) < 0)
    return 1;

  printf("CPM at 0x%lx...0x%lx.\n", cpm_base, cpm_base_end);

  cpm = jz4780_cpm_init(cpm_base);

  /* Read information from the clock and power management unit. */

  printf("Main source: %d\n", jz4780_cpm_get_source(cpm, Clock_main));
  printf("  Main frequency: %d\n", jz4780_cpm_get_frequency(cpm, Clock_main));
  printf("APLL frequency: %d\n", jz4780_cpm_get_frequency(cpm, Clock_pll_A));
  printf("EPLL frequency: %d\n", jz4780_cpm_get_frequency(cpm, Clock_pll_E));
  printf("MPLL frequency: %d\n", jz4780_cpm_get_frequency(cpm, Clock_pll_M));
  printf("VPLL frequency: %d\n", jz4780_cpm_get_frequency(cpm, Clock_pll_V));
  printf("CPU frequency: %d\n", jz4780_cpm_get_frequency(cpm, Clock_cpu));
  printf("Memory source: %d\n", jz4780_cpm_get_source(cpm, Clock_ddr));
  printf("  Memory source frequency: %d\n", jz4780_cpm_get_source_frequency(cpm, Clock_ddr));
  printf("  Memory frequency: %d\n", jz4780_cpm_get_frequency(cpm, Clock_ddr));
  printf("APB source: %d\n", jz4780_cpm_get_source(cpm, Clock_pclock));
  printf("  APB source frequency: %d\n", jz4780_cpm_get_source_frequency(cpm, Clock_pclock));
  printf("  Slow peripheral (APB) frequency: %d\n", jz4780_cpm_get_frequency(cpm, Clock_pclock));
  printf("AHB0 source: %d\n", jz4780_cpm_get_source(cpm, Clock_hclock0));
  printf("  AHB0 source frequency: %d\n", jz4780_cpm_get_source_frequency(cpm, Clock_hclock0));
  printf("  Fast peripheral (AHB0) frequency: %d\n", jz4780_cpm_get_frequency(cpm, Clock_hclock0));
  printf("AHB2 source: %d\n", jz4780_cpm_get_source(cpm, Clock_hclock2));
  printf("  AHB2 source frequency: %d\n", jz4780_cpm_get_source_frequency(cpm, Clock_hclock2));
  printf("  Fast peripheral (AHB2) frequency: %d\n", jz4780_cpm_get_frequency(cpm, Clock_hclock2));
  printf("LCD source: %d\n", jz4780_cpm_get_source(cpm, Clock_lcd));
  printf("  LCD source frequency: %d\n", jz4780_cpm_get_source_frequency(cpm, Clock_lcd));
  printf("  LCD frequency: %d\n", jz4780_cpm_get_frequency(cpm, Clock_lcd));
  printf("  LCD pixel clock frequency: %d\n", jz4780_cpm_get_frequency(cpm, Clock_lcd_pixel0));

  /* Attempt to set the pixel clock frequency. */

  jz4780_cpm_set_frequency(cpm, Clock_lcd_pixel0, 108000000);

  printf("LCD source: %d\n", jz4780_cpm_get_source(cpm, Clock_lcd));
  printf("  LCD source frequency: %d\n", jz4780_cpm_get_source_frequency(cpm, Clock_lcd));
  printf("  LCD frequency: %d\n", jz4780_cpm_get_frequency(cpm, Clock_lcd));
  printf("  LCD pixel clock frequency: %d\n", jz4780_cpm_get_frequency(cpm, Clock_lcd_pixel0));

  /* Attempt to set the peripheral clock frequency. */

  jz4780_cpm_set_source(cpm, Clock_pclock, 1);

  printf("APB source: %d\n", jz4780_cpm_get_source(cpm, Clock_pclock));
  printf("  APB source frequency: %d\n", jz4780_cpm_get_source_frequency(cpm, Clock_pclock));
  printf("  Slow peripheral (APB) frequency: %d\n", jz4780_cpm_get_frequency(cpm, Clock_pclock));
  printf("AHB2 source: %d\n", jz4780_cpm_get_source(cpm, Clock_hclock2));
  printf("  AHB2 source frequency: %d\n", jz4780_cpm_get_source_frequency(cpm, Clock_hclock2));
  printf("  Fast peripheral (AHB2) frequency: %d\n", jz4780_cpm_get_frequency(cpm, Clock_hclock2));

  /* Attempt to set the MPLL output to EXCLK divided by 4. */

  jz4780_cpm_set_parameters(cpm, Clock_pll_M, 3, (uint32_t []) {1, 4, 1});

  printf("APB source: %d\n", jz4780_cpm_get_source(cpm, Clock_pclock));
  printf("  APB source frequency: %d\n", jz4780_cpm_get_source_frequency(cpm, Clock_pclock));
  printf("  Slow peripheral (APB) frequency: %d\n", jz4780_cpm_get_frequency(cpm, Clock_pclock));
  printf("AHB2 source: %d\n", jz4780_cpm_get_source(cpm, Clock_hclock2));
  printf("  AHB2 source frequency: %d\n", jz4780_cpm_get_source_frequency(cpm, Clock_hclock2));
  printf("  Fast peripheral (AHB2) frequency: %d\n", jz4780_cpm_get_frequency(cpm, Clock_hclock2));

  return 0;
}
